@@ -75,6 +75,7 @@ def sqlite_s3_query(url, get_credentials=lambda: (
         if func(*args) != 0:
             raise Exception(libsqlite3.sqlite3_errmsg(db).decode())
 
+    @contextmanager
     def make_auth_request(http_client, method, params, headers):
         region, access_key_id, secret_access_key, session_token = get_credentials()
         to_auth_headers = headers + (
@@ -82,12 +83,12 @@ def sqlite_s3_query(url, get_credentials=lambda: (
             ()
         )
         request_headers = aws_sigv4_headers(
-            access_key_id, secret_access_key, region, 'GET', to_auth_headers, params,
+            access_key_id, secret_access_key, region, method, to_auth_headers, params,
         )
         url = f'{scheme}://{netloc}{path}?{urlencode(params)}'
-        response = http_client.get(url, headers=request_headers)
-        response.raise_for_status()
-        return response
+        with http_client.stream(method, url, headers=request_headers) as response:
+            response.raise_for_status()
+            yield response
 
     def aws_sigv4_headers(
         access_key_id, secret_access_key, region, method, to_auth_headers, params,
@@ -147,14 +148,26 @@ def sqlite_s3_query(url, get_credentials=lambda: (
 
     @contextmanager
     def get_vfs(http_client):
-        head_headers = make_auth_request(http_client, 'HEAD', (), ()).headers
+        with make_auth_request(http_client, 'HEAD', (), ()) as response:
+            head_headers = response.headers
         version_id = head_headers['x-amz-version-id']
         size = int(head_headers['content-length'])
-        get_range = lambda bytes_from, bytes_to: \
-            make_auth_request(http_client, 'HEAD',
-                (('versionId', version_id),),
-                (('range', f'bytes={bytes_from}-{bytes_to}'),)
-            ).content
+
+        def get_range(bytes_from, bytes_to):
+            with make_auth_request(http_client, 'GET',
+                    (('versionId', version_id),),
+                    (('range', f'bytes={bytes_from}-{bytes_to}'),)
+                ) as response:
+
+                # Handle the case of the server being broken or slightly evil, returning more than
+                # the number of bytes that's asked for
+                range_bytes = b''
+                for chunk in response.iter_bytes(chunk_size=bytes_to - bytes_from + 1):
+                    range_bytes += chunk
+                    if len(range_bytes) >= bytes_to - bytes_from + 1:
+                        break
+
+            return range_bytes
 
         def make_struct(fields):
             class Struct(Structure):
@@ -174,11 +187,15 @@ def sqlite_s3_query(url, get_credentials=lambda: (
         x_read_type = CFUNCTYPE(c_int, c_void_p, c_void_p, c_int, c_int64)
         def x_read(p_file, p_out, i_amt, i_ofst):
             try:
-                memmove(p_out, get_range(i_ofst, i_ofst + i_amt - 1), i_amt)
+                data = get_range(i_ofst, i_ofst + i_amt - 1)
             except Exception:
                 return SQLITE_IOERR_SHORT_READ
-            else:
-                return SQLITE_OK
+
+            if len(data) != i_amt:
+                return SQLITE_IOERR_SHORT_READ
+
+            memmove(p_out, data, i_amt)
+            return SQLITE_OK
 
         x_file_size_type = CFUNCTYPE(c_int, c_void_p, POINTER(c_int))
         def x_file_size(p_file, p_size):
